from . import BaseTool
import logging
import json
logger = logging.getLogger('sigmund')


class BaseOpenSesameTool(BaseTool):
    """Base class for OpenSesame tools that handles common command logic."""

    def __call__(self, **kwargs):
        """Generic call handler that builds command data from arguments."""
        command = self.__class__.__name__.replace('opensesame_', '')
        data = {"command": command}
        data.update(kwargs)
        # We don't ask for feedback directly, because OpenSesame will reply with
        # a user message.
        return ('(Suggesting OpenSesame action)', json.dumps(data, indent=2),
                'json', False)

    @property
    def required_arguments(self):
        """Automatically derive required arguments from arguments dict."""
        return list(self.arguments.keys())


class opensesame_add_existing_item_to_parent(BaseOpenSesameTool):
    """Adds an existing item to a parent item, which is typically a loop or a sequence. This creates a linked copy of the item."""

    arguments = {
        'item_name': {
            'type': 'string',
            'description': 'The name of the existing item to add.'
        },
        'parent_item_name': {
            'type': 'string',
            'description': 'The name of the parent item.'
        },
        'index': {
            'type': 'number',
            'description': 'The zero-based index where the item should be inserted in the parent.'
        }
    }


class opensesame_new_item(BaseOpenSesameTool):
    """Creates a new item and adds it to a parent item, which is typically a loop or a sequence. After the item has been created you will be able to define its script."""

    arguments = {
        'item_name': {
            'type': 'string',
            'description': 'The name of the new item to create.'
        },
        'item_type': {
            'type': 'string',
            'description': 'The type of the new item (e.g., "sketchpad", "keyboard_response", "inline_script").'
        },
        'parent_item_name': {
            'type': 'string',
            'description': 'The name of the parent item.'
        },
        'index': {
            'type': 'number',
            'description': 'The zero-based index where the item should be inserted in the parent.'
        }
    }


class opensesame_remove_item_from_parent(BaseOpenSesameTool):
    """Removes an item from a parent item, which is typically a loop or a sequence."""

    arguments = {
        'parent_item_name': {
            'type': 'string',
            'description': 'The name of the parent item.'
        },
        'index': {
            'type': 'number',
            'description': 'The zero-based index of the item to remove from the parent.'
        },       
    }


class opensesame_rename_item(BaseOpenSesameTool):
    """Renames an item in the experiment."""

    arguments = {
        'from_item_name': {
            'type': 'string',
            'description': 'The current name of the item.'
        },
        'to_item_name': {
            'type': 'string',
            'description': 'The new name for the item.'
        }
    }


class opensesame_select_item(BaseOpenSesameTool):
    """Selects an item in the OpenSesame interface. After selecting an item, you can modify its script."""

    arguments = {
        'item_name': {
            'type': 'string',
            'description': 'The name of the item to select.'
        }
    }


class opensesame_update_item_script(BaseOpenSesameTool):
    """Updates an item's script. You should only use this tool to modify the currently selected item."""

    arguments = {
        'item_name': {
            'type': 'string',
            'description': 'The name of the item (should be the currently selected item).'
        },
        'script': {
            'type': 'string',
            'description': 'The complete updated script for the item.'
        }
    }


class opensesame_set_global_var(BaseOpenSesameTool):
    """Sets a global experiment variable that can be accessed throughout the experiment."""

    arguments = {
        'var_name': {
            'type': 'string',
            'description': 'The name of the global variable.'
        },
        'value': {
            'type': ['string', 'number'],
            'description': 'The value to assign to the variable.'
        }
    }


from . import BaseTool
import logging
import json
from scholarly import scholarly
logger = logging.getLogger('sigmund')


class search_google_scholar(BaseTool):
    """Search Google Scholar for scientific articles. An overview of metadata (abstracts, authors, titles, etc.) of matching articles will be returned."""
    
    arguments = {
        "queries": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "description": "A list of search queries",
        }
    }
    required_arguments = ["queries"]
    
    def __call__(self, queries):
        results = []
        for query in queries:
            for i, result in enumerate(scholarly.search_pubs(query)):
                logger.info(f'appending doc for google scholar search: {query}')
                info = result['bib']
                if 'eprint_url' in result:
                    info['fulltext_url'] = result['eprint_url']
                results.append(info)
                if i >= 3:
                    break
        message = f'I found {len(results)} articles and added them to the workspace.'
        results = json.dumps(results, indent='  ')
        return message, results, True



import logging
from pathlib import Path
from datetime import datetime, timedelta
from flask import jsonify, redirect, Blueprint, url_for, request
from flask_login import login_required
from .. import config, utils
from ..database.manager import DatabaseManager
from .app import get_sigmund
import stripe
logger = logging.getLogger('sigmund')
subscribe_blueprint = Blueprint('subscribe', __name__)
stripe.api_key = config.stripe_secret_key


@subscribe_blueprint.route('/')
@login_required
def subscribe():
    """When the user is logged in but not subscribed, the request is redirected
    to the subscribe endpoint. This shows a simple page inviting the user to
    subscribe through Stripe. If the user is already subscribed (which can 
    happen if the endpoint is accessed directly) then the user is redirected
    to the Stripe customer portal.
    """
    if not config.subscription_required:
        return redirect(url_for('app.chat'), code=303)    
    sigmund = get_sigmund()
    if sigmund.database.check_subscription():
        logger.info(f'redirecting {sigmund.user_id} to customer portal')
        redirect(url_for('subscribe.customer_portal'), code=303)
    username = sigmund.user_id
    if '(google)::' in username:
        username = username.split('(google)::')[0] + '(Google)'
    return utils.render('subscribe-now.html', username=username)
    
    
@subscribe_blueprint.route('/create-checkout-session')
@login_required
def create_checkout_session():
    """Prepares the product for checkout and then redirects to Stripe for the
    actual transaction.
    """
    sigmund = get_sigmund()
    logger.info(f'initiating checkout for {sigmund.user_id}')
    try:
        checkout_session = stripe.checkout.Session.create(
            success_url=f'{config.server_url}/subscribe/success/{{CHECKOUT_SESSION_ID}}',
            cancel_url=f'{config.server_url}/',
            client_reference_id=sigmund.user_id,
            mode='subscription',
            line_items=[{'price': config.stripe_price_id, 'quantity': 1}],
            **config.stripe_checkout_keywords)
        return redirect(checkout_session.url, code=303)
    except Exception as e:
        return jsonify({'error': {'message': str(e)}}), 400


@subscribe_blueprint.route('/success/<checkout_session_id>')
@login_required
def success(checkout_session_id):
    """This is called by Stripe after a subscription was succesfully finished.
    The actual subscription is activated in the webhook. This happens before
    the success endpoint is called.
    
    An error should not occur at this point. If it does, we show an error page
    inviting the user to try again and offer contact information.
    """
    sigmund = get_sigmund()
    try:
        checkout_session = stripe.checkout.Session.retrieve(
            checkout_session_id)
    except stripe.error.StripeError as e:
        logger.error(f'Stripe API error ({checkout_session_id}): {e}')
        return utils.render('subscribe-error.html'), 500
    return utils.render('subscribe-success.html')


@subscribe_blueprint.route('/customer-portal')
@login_required
def customer_portal():
    """The customer portal redirects to Stripe so that the user can manage
    subscription information. This is only valid for users with a current or
    previous subscription.
    
    An error should not occur at this point. If it does, we show an error page
    inviting the user to try again and offer contact information.
    """
    sigmund = get_sigmund()
    customer_id = sigmund.database.get_stripe_customer_id()
    if not customer_id:
        logger.error('No Stripe customer ID found for the user')
        return utils.render('subscribe-error.html'), 404
    try:
        session = stripe.billing_portal.Session.create(
            customer=customer_id,
            return_url=config.server_url)
        return redirect(session.url, code=303)
    except stripe.error.StripeError as e:
        logger.error(f'Stripe API error: {e}')
        return utils.render('subscribe-error.html'), 500


@subscribe_blueprint.route('/cancel')
@login_required
def cancel():
    """This entrypoint is for internal use and allows a subscription to be
    canceled independently of Stripe. Normally, a subscription would be
    canceled through the webhook.
    """
    sigmund = get_sigmund()
    sigmund.database.cancel_subscription()
    return redirect(url_for('subscribe.subscribe'), code=303)


@subscribe_blueprint.route('/webhook', methods=['POST'])
def webhook():
    """The webhook is the main mechanism through which Stripe informs the app
    of payments. The webhook is a public entry point, which means that the
    user is not logged in.
    """
    payload = request.data
    sig_header = request.headers.get('Stripe-Signature')
    event = None
    try:
        event = stripe.Webhook.construct_event(
            payload, sig_header, config.stripe_webhook_secret)
    except ValueError as e:
        logger.error(f'invalid webhook payload: {e}')
        return 'Invalid payload', 400
    except stripe.error.SignatureVerificationError as e:
        logger.error(f'invalid webhook signature: {e}')
        return 'Invalid signature', 400
    if not event:
        logger.error('webhook contains no event')
        return jsonify(success=True), 200
    event_type = event['type']
    if event_type not in ('checkout.session.completed',
                          'invoice.payment_succeeded'):
        logger.error(f'webhook ignored: {event_type}')
        return jsonify(success=True), 200
    event_object = event['data']['object']
    stripe_customer_id = event_object['customer']
    stripe_subscription_id = event_object.get('subscription', None)
    logger.info(f'webhook event: {event_type}')
    # checkout.session.completed is sent when the user completes a checkout to
    # start a new subscription. At this point we associate the sigmund user id
    # with the stripe customer and subscription ids. This event is not sent
    # for subscription renewals, which are handled below.
    if event_type == 'checkout.session.completed':
        sigmund_user_id = event_object['client_reference_id']
        database = DatabaseManager(None, sigmund_user_id)
        database.update_subscription(
            stripe_customer_id=stripe_customer_id,
            stripe_subscription_id=stripe_subscription_id)
        logger.info(
            f'completed checkout for {sigmund_user_id} '
            f'(stripe_customer_id: {stripe_customer_id}, '
            f'stripe_subscription_id: {stripe_subscription_id})')                        
    # invoice.payment_succeeded is sent when a payment was successfully 
    # completed, both for new subscriptions and renewals.
    elif event_type == 'invoice.payment_succeeded':
        # Attempt to get the user-specific database instance based on the
        # stripe customer id
        database = DatabaseManager.from_stripe_customer_id(stripe_customer_id)
        # For new subscriptions, this fails because the link between the 
        # stripe customer id and the sigmund user id still needs to be 
        # established in checkout.session.completed, which is fired later.
        if database is None:
            logger.info('subscription doesn\"t exist yet, waiting '
                        'for checkout.session.completed')
        # For renewals, this succeeds and we renew the subscription
        else:
            database.update_subscription(
                stripe_customer_id=stripe_customer_id,
                stripe_subscription_id=stripe_subscription_id)
            logger.info(
                f'received payment for {database.username} '
                f'(stripe_customer_id: {stripe_customer_id}, '
                f'stripe_subscription_id: {stripe_subscription_id})')
    logger.info('webhook successful')
    return jsonify(success=True), 200



from ._base_tool import BaseTool
from ._search_google_scholar import search_google_scholar
from ._execute_code import execute_code
from ._generate_image import generate_image
from ._opensesame_tools import (opensesame_add_existing_item_to_parent,
                                opensesame_new_item,
                                opensesame_remove_item_from_parent,
                                opensesame_rename_item,
                                opensesame_select_item,
                                opensesame_set_global_var,
                                opensesame_update_item_script)



import logging
import asyncio
import time
import re
logger = logging.getLogger('sigmund')


class BaseModel:
    """Base implementation for LLM chat models."""
    
    # Indicates whether the model is able to provide feedback on its own output
    supports_not_done_yet = False
    # Indicates whether the model is able to provide feedback on tool results 
    supports_tool_feedback = True
    # Approximation to keep track of token costs
    characters_per_token = 4
    
    def __init__(self, sigmund, model, thinking=False, tools=None,
                 tool_choice='auto'):
        self._sigmund = sigmund
        self._model = model
        self._thinking = thinking
        self._tools = tools
        self._tool_choice = tool_choice
        self.total_tokens_consumed = 0
        self.prompt_tokens_consumed = 0
        self.completion_tokens_consumed = 0
        self.json_mode = False
        
    def invalid_tool(self) -> str:
        return 'Invalid tool', None, 'markdown', False
        
    def get_response(self, response) -> [str, callable]:
        return response.content
    
    def tools(self):
        if self._tools is None:
            return []
        return [{"type": "function", "function": t.tool_spec}
                for t in self._tools if t.tool_spec]
        
    def invoke(self, messages, attachments=None):
        raise NotImplementedError()
        
    def async_invoke(self, messages, attachments=None):
        raise NotImplementedError()
        
    def messages_length(self, messages) -> int:
        if isinstance(messages, str):
            return len(messages)
        return sum([len(m.content if hasattr(m, 'content') else m['content'])
                   for m in messages])
        
    def convert_message(self, message):
        if isinstance(message, str):
            return dict(role='user', content=message)
        if isinstance(message, dict):
            return message
        raise ValueError(f'Unknown message type: {message}')

    def predict(self, messages, attachments=None, track_tokens=True):
        t0 = time.time()
        logger.info(f'predicting with {self.__class__} model')
        reply = self.get_response(self.invoke(messages))
        msg_len = self.messages_length(messages)
        dt = time.time() - t0
        prompt_tokens = msg_len // self.characters_per_token
        reply_len = len(reply) if isinstance(reply, str) else 0
        logger.info(f'predicting {reply_len + msg_len} took {dt} s')
        if track_tokens:
            completion_tokens = reply_len // self.characters_per_token
            total_tokens = prompt_tokens + completion_tokens
            self.total_tokens_consumed += total_tokens
            self.prompt_tokens_consumed += prompt_tokens
            self.completion_tokens_consumed += completion_tokens
            logger.info(f'total tokens (approx.): {total_tokens}')
            logger.info(f'prompt tokens (approx.): {prompt_tokens}')
            logger.info(f'completion tokens (approx.): {completion_tokens}')
        return reply
    
    def predict_multiple(self, prompts):
        """Predicts multiple simple (non-message history) prompts using asyncio
        if possible.
        """
        prompts = [[self.convert_message(prompt)] for prompt in prompts]
        try:
            loop = asyncio.get_event_loop()
            if not loop.is_running():
                logger.info('re-using async event loop')
                use_async = True
            else:
                logger.info('async event loop is already running')
                use_async = False
        except RuntimeError:
            logger.info('creating async event loop')
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            use_async = True
            
        if not use_async:
            logger.info('predicting multiple without async')
            return [self.get_response(self.invoke(prompt))
                                              for prompt in prompts]
            
        async def wrap_gather():
            tasks = [self.async_invoke(prompt) for prompt in prompts]
            try:
                predictions = await asyncio.gather(*tasks)
            except Exception as e:
                logger.warning(f'failed to gather predictions: {e}')
                return []
            return [self.get_response(p) for p in predictions]
            
        logger.info('predicting multiple using async')
        return loop.run_until_complete(wrap_gather())

    @staticmethod
    def embed_thinking_block(signature: str | None, content: str | None) -> str:
        """Embeds information about an Anthropic thinking block as an HTML element."""
        sig = f'<div class="thinking_block_signature">{signature}</div>' if signature else ""
        cont = f'<div class="thinking_block_content">{content}</div>' if content else ""
        return sig + cont
    
    @staticmethod
    def extract_thinking_block(content: str) -> [str, str | None, str | None]:
        """
        Extracts signature/content spans if present and returns a tuple:
          (cleaned_content, signature, content)
        """
        sig_pattern = r'<div\s+class="thinking_block_signature">(.*?)</div>'
        cont_pattern = r'<div\s+class="thinking_block_content">(.*?)</div>'
        signature = None
        thinking_content = None
    
        sig_match = re.search(sig_pattern, content)
        if sig_match:
            signature = sig_match.group(1)
            content = re.sub(sig_pattern, '', content, count=1)
    
        cont_match = re.search(cont_pattern, content, re.MULTILINE | re.DOTALL)
        if cont_match:
            thinking_content = cont_match.group(1)
            content = re.sub(cont_pattern, '', content, count=1,
                             flags=re.MULTILINE | re.DOTALL)
    
        cleaned = content.strip()
        if signature or thinking_content:
            return cleaned, signature, thinking_content
        return content, None, None


from . import BaseModel
from .. import config, utils
import logging
import json
logger = logging.getLogger('sigmund')


class AnthropicModel(BaseModel):
    
    supports_not_done_yet = False
    
    def __init__(self, sigmund, model, **kwargs):
        from anthropic import Anthropic, AsyncAnthropic
        super().__init__(sigmund, model, **kwargs)
        self._tool_use_id = 0
        self._client = Anthropic(api_key=config.anthropic_api_key)
        self._async_client = AsyncAnthropic(api_key=config.anthropic_api_key)
        
    def predict(self, messages, attachments=None, track_tokens=True):
        # import pprint
        # print('=== preparing tool messages')
        # pprint.pprint(messages)
        # print('---')        
        if isinstance(messages, str):
            return super().predict([self.convert_message(messages)],
                                   attachments, track_tokens)
        messages = utils.prepare_messages(messages, allow_ai_first=False,
                                          allow_ai_last=False,
                                          merge_consecutive=True)
        messages = [self.convert_message(message) for message in messages]
        # The Anthropic messages API expects thinking blocks if thinking mode is
        # enabled. These thinking blocks are embedded in plain text in the 
        # responses and should be extracted. This also changes the message 
        # content from plain text to a list of blocks.
        if self._thinking:
            for message in messages:
                # only AI messages have thinking blocks
                if message['role'] != 'assistant':
                    continue                
                content = message['content']
                content, thinking_signature, thinking_content = \
                    self.extract_thinking_block(content)
                # It can happen that the response only contains thinking blocks
                # and workspace content, resulting in empty regular content.
                # Empty text blocks are not allowed, so in that case we add this
                # filler message.
                if not content.strip():
                    content = 'I added content to the workspace.'
                if thinking_signature is None:
                    continue
                message['content'] = [
                    {'type': 'thinking',
                     'thinking': thinking_content,
                     'signature': thinking_signature},
                    {'type': 'text',
                     'text': content}
                ]
                logger.info('extracted thinking block')
        # The Anthropic messages API doesn't accept tool results in a separate
        # message. Instead, tool results are included as a special content 
        # block in a user message. Since two subsequent user messages aren't
        # allowed, we need to convert a tool message to a user message and if
        # necessary merge it with the next user message.
        # import pprint
        # print('*** after extracting thinking block')
        # pprint.pprint(messages)
        # print('***')
        while True:
            logger.info('entering message postprocessing loop')
            for i, message in enumerate(messages):
                if message['role'] == 'tool':
                    if i == 0:
                        raise ValueError(
                            'The first message cannot be a tool message')
                    logger.info('converting tool message to user message')
                    tool_info = json.loads(message['content'])
                    message['role'] = 'user'
                    if tool_info['content'] is None:
                        tool_info['content'] = ''
                    message['content'] = [{
                        'type': 'tool_result',
                        # IDs formatted like this: toolu_01WaeSyitUGJFaaPe68cJuEv
                        'tool_use_id': f'toolu_{self._tool_use_id:024}',
                        'content': [{
                            'type': 'text',
                            'text': tool_info['content']
                        }]
                    }]
                    # The previous message needs to have a tool-use block. If
                    # necessary, this means that the content is changed to a
                    # list of blocks from a plain text.
                    prev_message = messages[i - 1]
                    if isinstance(prev_message['content'], str):                        
                        prev_message['content'] = [
                            {'type': 'text',
                             'text': prev_message['content']}
                        ]
                    prev_message['content'].append(
                        {'type': 'tool_use',
                         'id': f'toolu_{self._tool_use_id:024}',
                         'input': {'args': tool_info['args']},
                         'name': tool_info['name']
                        }
                    )
                    self._tool_use_id += 1
                    if len(messages) > i + 1:
                        next_message = messages[i + 1]
                        if next_message['role'] == 'user':
                            logger.info('merging tool and user message')
                            message['content'].append({
                                "type": "text",
                                "text": next_message['content']
                            })
                            break
            else:
                break
            logger.info('dropping duplicate user message')
            messages.remove(next_message)
        # The first message is used as the system message. This has to be a 
        # text-only, which means that any tool_use block needs to be removed.
        # By extension, this means that any tool_result block in the subsequent
        # user message should also be removed, because a tool_result block needs
        # to be preceded by a tool_use block.
        if isinstance(messages[0]['content'], list):
            logger.info('removing tool use and result from first two messages')
            messages[0]['content'] = messages[0]['content'][0]['text']
            messages[1]['content'] = messages[1]['content'][0]['content'][0]['text']
        # print('*** after tool processing')
        # pprint.pprint(messages)
        # print('***')            
        # Attachments are included with the last message. The content is now
        # no longer a single str, but a list of dict
        if attachments:
            logger.info('adding attachments to last message')
            content = messages[-1]['content']
            # If the content is plain text, convert it to a list of blocks
            if isinstance(content, str):
                content = [{'type': 'text', 'text': content}]
            for attachment in attachments:
                # Decompose the HTML-style data into a mimetype and the 
                # actual data
                url = attachment['url']
                mimetype = url[5:url.find(';')]
                data = url[url.find(',') + 1:]
                if attachment['type'] == 'image':
                    content.append({
                        'type': 'image',
                        'source': {'type': 'base64',
                                   'media_type': mimetype,
                                   'data': data}
                    })
                elif attachment['type'] == 'document':
                    content.append({
                        'type': 'document',
                        'source': {'type': 'base64',
                                   'media_type': mimetype,
                                   'data': data}
                    })
            messages[-1]['content'] = content
        # print('*** after attachment processing')
        # import pprint
        # pprint.pprint(messages)
        # print('***')            
        return super().predict(messages, attachments, track_tokens)
        
    def get_response(self, response):
        text = []
        tool_message_prefix = ''
        thinking_block = None
        for block in response.content:
            if block.type == 'tool_use':
                if self._tools:
                    for tool in self._tools:
                        if tool.name == block.name:
                            return tool.bind(
                                json.dumps(block.input),
                                message_prefix=tool_message_prefix + '\n\n')
                return self.invalid_tool
            if block.type == 'text':
                text.append(block.text)
                tool_message_prefix += block.text
            if block.type == 'thinking':
                thinking_block = self.embed_thinking_block(block.signature,
                                                           block.thinking)
                tool_message_prefix += thinking_block
        if thinking_block is not None:
            text.insert(0, thinking_block)
        return '\n'.join(text)
        
    def _tool_args(self):
        if not self._tools:
            return {}
        alternative_format_tools = []
        for tool in self.tools():
            if tool['type'] == 'function':
                function = tool['function']
                alt_tool = {
                    "name": function['name'],
                    "description": function['description'],
                    "input_schema": function['parameters']
                }
                alternative_format_tools.append(alt_tool)
        return {'tools': alternative_format_tools}
        
    def _anthropic_invoke(self, fnc, messages):
        kwargs = self._tool_args()
        kwargs.update(config.anthropic_kwargs)
        # If the first message is the system prompt, we need to separate this
        # from the user and assistant messages, because the Anthropic messages
        # API takes this as a separate keyword argument.
        if messages[0]['role'] == 'system':
            kwargs['system'] = messages[0]['content']
            messages = messages[1:]
        if self._thinking:
            kwargs['thinking'] = {
                "type": "enabled",
                "budget_tokens": config.anthropic_max_thinking_tokens
            }
        try:
            return fnc(model=self._model, messages=messages, **kwargs)
        except Exception:            
            import pprint
            print('=== an error occurred while sending messages')
            pprint.pprint(messages)
            print('=== system message')
            print(kwargs['system'])
            print('***')
            raise
        
    def invoke(self, messages):
        return self._anthropic_invoke(self._client.messages.create, messages)
        
    def async_invoke(self, messages):
        return self._anthropic_invoke(self._async_client.messages.create,
                                      messages)
